// ==============
// File: my_string.cpp
// By: Jack Holdsworth - 20023933
// Use: MyString class implementation for task2
// Written on: c++17
// ==============

#include "my_string.hpp"

// Empty constructor
MyString::MyString() {
    string = nullptr; // Initialize the string pointer to nullptr
}

// Constructor
MyString::MyString(const char* strVar) {
    // Allocate memory for the reference count & string (+1 for null terminator )
    count = new int;
    string = new char[std::strlen(strVar) + 1];

    //Copy into allocated memory and set count for new instance 
    strcpy(string, strVar);
    *count = 1;
}

// Copy constructor
MyString::MyString(const MyString& strVar) {
    std::cout << "Copy constructor" << std::endl;
    
    //Copy and increment 
    string = strVar.string;
    count = strVar.count;
    ++*count;
}

// Assignment operator
MyString& MyString::operator=(const MyString& strVar) {
    std::cout << "Assignment constructor " << std::endl;
    
    if (this != &strVar) { // Check if source and target objects are different
        if (count != nullptr && --(*count) == 0) { // Decrement the reference count and check if it's zero
            //Delete the allocated string & reference memory
            delete[] string;
            delete count;
        }

        // Copy pointers and inc reference count for the new inst
        string = strVar.string; 
        count = strVar.count; 
        *count = *count + 1; 
    }

    return *this; 
}

// Destructor
MyString::~MyString() {
    std::cout << "Destructor " << std::endl;

    if (count != nullptr) {
        --*count;

        //If the reference counter reaches zero ( meaning no more references) then delete and free memory.
        if (*count <= 0) { 

            // Delete the allocated string & count when an instance is destroyed
            delete[] string;
            delete count;
        }
    }
}


// Get character at a specific position
char MyString::getChar(const int& iPos) const {
    return string[iPos]; 
}

// Set character at a specific position
void MyString::setChar(const int& iPos, const char& charVar) {
    string[iPos] = charVar;
}

// Print the string and its reference count
void MyString::print() const {
    std::cout << string << " [" << *count << ']' << std::endl; 
}